#! /usr/bin/env python
#


""" Library to convert 'meta' information into URL assuming: 
https://irsa.ipac.caltech.edu/docs/program_interface/ztf_metadata.html
"""
import os
import warnings
import numpy as np
# ========================== #
#                            #
#   Generic Query            #
#                            #
# ========================== #

KNOWN_SCIENCE_SUFFIXES = {    "sciimg.fits":"(primary science image)",
                      "mskimg.fits":"(bit-mask image)",
                      "psfcat.fits":"(PSF-fit photometry catalog)",
                      "sexcat.fits":"(nested-aperture photometry catalog)",
                      "sciimgdao.psf":"(spatially varying PSF estimate in DAOPhot's lookup table format)",
                      "sciimgdaopsfcent.fits":"(PSF estimate at science image center as a FITS image)",
                      "sciimlog.txt":"(log output from instrumental calibration pipeline)",
                      "scimrefdiffimg.fits.fz":"(difference image: science minus reference; fpack-compressed)",
                      "diffimgpsf.fits":"(PSF estimate for difference image as a FITS image)",
                      "diffimlog.txt":"(log output from image subtraction and extraction pipeline)",
                      "log.txt":"(overall system summary log from realtime pipeline)"}

# SOURCES 
DATA_BASEURL   = "https://irsa.ipac.caltech.edu/ibe/data/ztf/products/"
ZTFIRSA_BASE = ["/ref/","/sci/","/raw/","/cal/"]

from .io import LOCALSOURCE
DEFAULT_SOURCE = DATA_BASEURL


# ================== #
#  Building the URL  #
# ================== #
def _localsource_to_source_(localfilepath, source=None):
    """ """
    filesource = _source_to_location_(source)
    if LOCALSOURCE not in localfilepath:
        raise ValueError("Cannot parse the location of the given file. %s, %s expected to be part of the path"%(localfilepath,LOCALSOURCE))

    if np.any([k in localfilepath for k in ZTFIRSA_BASE]):
        return localfilepath.replace(LOCALSOURCE, filesource), "irsa"
    else:
        warnings.warn("Cannot recover the online url, only IRSA ZTF files are implemented.")
        return None, "None"
    
def _source_to_location_(source):
    """ convert flexible source naming into actual source path """
    if source is None: source = "irsa"
    source = source.lower()

    if source in ["none"]:
        return ""
    
    if source in [DATA_BASEURL, LOCALSOURCE] or "/" in source:
        return source
    
    if source in ['irsa', 'ipac', "web", "online", "ztf"]:
        return DEFAULT_SOURCE
        
    if source in ["local", "cmp", "computer"]:
        return LOCALSOURCE
    
    if source in ['nersc']:
        raise ValueError("NERCS sourcing not ready yet.")
    
    raise ValueError("Cannot parse the given source %s"%source)


# ------------- #
#  Sci & Raw    #
# ------------- #
def science_path(year, month, day, fracday,
                paddedfield,
                filtercode, paddedccdid, qid,
                imgtypecode="o", suffix="sciimg.fits",
                source="", verbose=False):
    """ 
    suffix: [string]
        What kind of data do you want?
        - sciimg.fits (primary science image) [used if suffix is None]
        - mskimg.fits (bit-mask image)
        - psfcat.fits (PSF-fit photometry catalog)
        - sexcat.fits (nested-aperture photometry catalog)
        - sciimgdao.psf (spatially varying PSF estimate in DAOPhot's lookup table format)
        - sciimgdaopsfcent.fits (PSF estimate at science image center as a FITS image)
        - sciimlog.txt (log output from instrumental calibration pipeline)
        - scimrefdiffimg.fits.fz (difference image: science minus reference; fpack-compressed)
        - diffimgpsf.fits (PSF estimate for difference image as a FITS image)
        - diffimlog.txt (log output from image subtraction and extraction pipeline)
        - log.txt (overall system summary log from realtime pipeline)
        
    filtercode: [2 digit string]
        The filter name ('zg','zr','zi', or 'OO' for Bias)
        
    paddedccdid: [2 digit string]
        The ccd id [01,02,03....15,16]

    qid: [1 digit string]
        Which quadran [1,2,3,4]
    
    """
    if verbose:
        print(locals())
    source = _source_to_location_(source)
    if suffix is None:
        suffix="sciimg.fits"
    elif suffix not in KNOWN_SCIENCE_SUFFIXES.keys():
        raise ValueError("Unkwown suffix %s for 'sci' exposures"%suffix, "\n known suffixes: \n", KNOWN_SCIENCE_SUFFIXES)
    
    filefracday = "".join([year+month+day+fracday])
    return source+'sci/'+year+'/'+month+day+'/'+fracday+'/ztf_'+filefracday+'_'+paddedfield+'_'+filtercode+'_c'+paddedccdid+'_'+imgtypecode+'_q'+qid+'_'+suffix
    
def raw_path(year, month, day, fracday,
            paddedfield,
            filtercode, paddedccdid, imgtypecode="o",
            source=""):
    """ 

    Parameters
    ----------
    filtercode: [2 digit string]
        The filter name ('zg','zr','zi', or 'OO' for Bias)
        
    paddedccdid: [2 digit string]
        The ccd id [01,02,03....15,16]

    imgtypecode: [char]
        Currently, the possible values for "imgtypecode" pertaining to raw image data files are:
        - o = science (on-sky); see example above.
        - b = bias calibration image
        - d = dark calibration image
        - f = dome/screen flatfield calibration image

    """
    source = _source_to_location_(source)
    filefracday = "".join([year+month+day+fracday])
    return source+'raw/'+year+'/'+month+day+'/'+fracday+'/ztf_'+filefracday+'_'+paddedfield+'_'+filtercode+'_c'+paddedccdid+'_'+imgtypecode+'.fits.fz'

# ------------- #
#  Calibration  #
# ------------- #
def calibration_path(caltype,
                    year, month, day,
                    filtercode,
                    paddedccdid, qid, suffix=None,
                    source=""):
    """ 
    Parameters:
    -----------
    caltype: [string]
        Should be 'bias' or 'hifreqflat'
    
    filtercode: [2 digit string]
        The filter name ('zg','zr','zi', or 'OO' for Bias)
        
    paddedccdid: [2 digit string]
        The ccd id [01,02,03....15,16]

    qid: [1 digit string]
        Caltype quadran [1,2,3,4]

    suffix: [string] -optional-
        What kind of calibration file you want for the given `caltype`
        - None: returns `caltype`.fits
        - log:  returns `caltype`log.txt
        - unc:  returns `caltype`unc.fits
    """
    # ================ #
    #  INPUT TESTING   #
    # ================ #
    source = _source_to_location_(source)
    caltype = caltype.lower() # case insensitive
    if caltype.lower() not in ["bias", "hifreqflat"]:
        raise ValueError("Unknown `caltype` %s"%caltype+" caltype must be  'bias' or 'hifreqflat'")

    if suffix is None or suffix in ["basic",".fits","fits"]:
        suffix = caltype+".fits"
    elif suffix in ["log", "log.txt"]:
        suffix = caltype+"log.txt"
    elif suffix in ["unc", "uncorrected","unc.fits"]:
        suffix = caltype+"unc.fits"
    else:
        raise ValueError("Unknow suffix %s must be: basic/None or log or unc")
    
    # ================ #
    #  Build URL       #
    # ================ #
    filestartdate = "".join([year,month,day])
    return source+'cal/'+year+'/'+month+day+'/%s/'%caltype.lower()+filtercode+'/ccd'+paddedccdid+'/q'+qid+'/ztf_'+filestartdate+'_'+filtercode+'_c'+paddedccdid+'_q'+qid+'_%s'%suffix
    

# ------------- #
#  References   #
# ------------- #
def reference_path(paddedfield,
                  filtercode,
                  paddedccdid, qid,
                  fieldprefix="000",
                  suffix=None,
                  source="", verbose=True):
    """   

    filtercode: [2 digit string]
        The filter name ('zg','zr','zi', or 'OO' for Bias)
        
    paddedccdid: [2 digit string]
        The ccd id [01,02,03....15,16]

    qid: [1 digit string]
        Which quadran [1,2,3,4]

    suffix: [string]
        Could be:
        - log.txt
        - refcov.fits
        - refimg.fits
        - refimlog.txt
        - refpsfcat.fits
        - refsexcat.fits
        - refunc.fits

    """
    if verbose:
        print(locals())

    if suffix is None:
        suffix = "refimg.fits"
    source = _source_to_location_(source)
    
    return source+'ref/'+fieldprefix+'/field'+paddedfield+'/'+filtercode+'/ccd'+paddedccdid+'/q'+qid+'/ztf_'+paddedfield+'_'+filtercode+'_c'+paddedccdid+'_q'+qid+'_%s'%suffix

# ============= #
#   TOOLS       #
# ============= #
def filename_to_scienceurl(filename, suffix=None, source="irsa", verbose=False):
    """ 
    """
    _, filefracday, paddedfield, filtercode, ccd_, imgtypecode, qid_, suffix_ = os.path.basename(filename).split("_")
    year,month, day, fracday = filefrac_to_year_monthday_fracday(filefracday)
    paddedccdid = ccd_.replace("c","")
    qid = qid_.replace("q","")
    
    if suffix is None:
        suffix = suffix_
        
    return science_path(year, month, day, fracday,
                paddedfield, filtercode,
                paddedccdid, qid, # added in input
                imgtypecode=imgtypecode, suffix=suffix,
                source=source, verbose=verbose)
    

def filename_to_refurl(filename, suffix, source="irsa", verbose=False):
    """ 
    suffix: [string]
        Could be:
        - log.txt
        - refcov.fits
        - refimg.fits
        - refimlog.txt
        - refpsfcat.fits
        - refsexcat.fits
        - refunc.fits
    """
    _, filefracday, paddedfield, filtercode, ccd_, imgtypecode, qid_, suffix_ = os.path.basename(filename).split("_")
    year,month, day, fracday = filefrac_to_year_monthday_fracday(filefracday)
    paddedccdid = ccd_.replace("c","")
    qid = qid_.replace("q","")
    
    
    return reference_path(paddedfield,
                                   filtercode,
                                   paddedccdid, qid,
                                   fieldprefix=paddedfield[:3],
                                   suffix=suffix,
                                   source=source)


def filefrac_to_year_monthday_fracday(filefracday):
    """ split the name as YYYY, MM,DD, FFFFFF
    FFFFFF (fraction in the day)"""
    return filefracday[:4], filefracday[4:6], filefracday[6:8],filefracday[8:]


def fileroot_to_science_url(fileroot, paddedccdid, qid,
                            imgtypecode="o", suffix="sciimg.fits", source="",
                            verbose=False):
    """ 
    Parameters
    ----------

    """
    if suffix not in KNOWN_SCIENCE_SUFFIXES.keys():
        raise ValueError("Unkwown suffix %s for 'sci' exposures"%suffix , "\n known suffixes: \n", KNOWN_SCIENCE_SUFFIXES)
    
    _, filefracday, paddedfield, filtercode = fileroot.split("_")
    year,month, day, fracday = filefrac_to_year_monthday_fracday(filefracday)
    return science_path(year, month, day, fracday,
                paddedfield, filtercode,
                paddedccdid, qid, # added in input
                imgtypecode=imgtypecode, suffix=suffix,
                source=source, verbose=verbose)


def fileroot_to_raw_url(fileroot, paddedccdid,
                        imgtypecode="o", source="", verbose=False):
    """ 
    Parameters
    ----------

    """
    _, filefracday, paddedfield, filtercode = fileroot.split("_")
    year, month, day, fracday = filefrac_to_year_monthday_fracday(filefracday)

    return raw_path(year, month, day, fracday,
            paddedfield,
            filtercode, paddedccdid, imgtypecode=imgtypecode,
            source=source)
    




""" Generic projection and plotting tools """

import numpy as np

_DEG2RA = np.pi / 180



##############################
#                            #
#  MPL Tools                 #
#                            #
##############################
def colorbar(ax,cmap,vmin=0,vmax=1,label="",
             fontsize="x-large",**kwargs):
    """ Set a colorbar in the given axis

    Parameters
    -----------
    ax: [mpl's Axes]
        Axis in which the colorbar will be drawn

    cmap: [mpl's colormap]
        A matplotlib colormap

    vmin, vmax: [float,float] -optional-
        Extend of the colormap, values of the upper and lower colors

    label, fontsize: [string, string/float] -optional-
        Label of the colorbar and its associated size
     
    **kwargs goes to matplotlib.colobar.ColorbarBase

    Return
    ------
    colorbar
    """
    import matplotlib
    
    if "orientation" not in kwargs.keys():
        bbox = ax.get_position()
        orientiation = "vertical" if bbox.xmax - bbox.xmin < bbox.ymax - bbox.ymin \
          else "horizontal"
        kwargs["orientation"] = orientiation

    norm    = matplotlib.colors.Normalize(vmin=vmin,vmax=vmax)
    c_bar   = matplotlib.colorbar.ColorbarBase(ax, cmap=cmap,
                              norm=norm,**kwargs)
    
    c_bar.set_label(label,fontsize=fontsize)
    if "ticks" in kwargs.keys() and "ticklabels" not in kwargs.keys():
        c_bar.ax.set_xticklabels([r"%s"%v for v in kwargs["ticks"]])
        
    return c_bar

def insert_ax(ax, location, shrunk=0.7, space=.05,
              axspace=0.02, shareax=False, **kwargs):
    """ insert an axis at the requested location

              
    The new axis will share the main axis x-axis (location=top or bottom) or
    the y-axis (location=left or right).

    Parameters:
    -----------
    location: [string]
       top/bottom/left/right, i.e. where new axis will be set

    shrunk: [float]
        the main axis will be reduced by so much (0.7 = 70%).
        the new axis will take the room

    space: [float]
        extra space new axis does not use between it and the edge of
        the figure. (in figure unit, i.e., [0,1])

    axspace: [float]
        extra space new axis does not use between it and the input
        axis. (in figure unit, i.e., [0,1])

    shareax: [bool]
        The new axis will share the main axis x-axis (location=top or bottom) or
        the y-axis (location=left or right). If so, the axis ticks will be cleaned.
                           
    **kwargs goes to figure.add_axes() for the new axis

    Returns:
    --------
    axes (the new axis)
    """
    from matplotlib.transforms  import Bbox
    # --------------------
    # hist x
    # -------------------- #
    # -- keep trace of the original axes
    bboxorig = ax.get_position().frozen()

    if location in ["top","bottom"]:
        axhist = ax.figure.add_axes([0.1,0.2,0.3,0.4],sharex=ax if shareax else None,
                                    **kwargs) # This will be changed
        _bboxax = ax.get_position().shrunk(1,shrunk)
        _bboxhist = Bbox([[_bboxax.xmin, _bboxax.ymax+axspace ],
                          [_bboxax.xmax, bboxorig.ymax-space]])
        
        if location == "bottom":
            tanslate = _bboxhist.height + space+axspace
            _bboxhist = _bboxhist.translated(0, bboxorig.ymin-_bboxhist.ymin+space)
            _bboxax = _bboxax.translated(0,tanslate)
            
    # --------------------
    # hist y
    # -------------------- #            
    elif location in ["right","left"]:
        axhist = ax.figure.add_axes([0.5,0.1,0.2,0.42],sharey=ax if shareax else None,
                                    **kwargs) # This will be changed
        _bboxax = ax.get_position().shrunk(shrunk,1)
        _bboxhist = Bbox([[_bboxax.xmax+axspace, _bboxax.ymin ],
                          [bboxorig.xmax-space, _bboxax.ymax]])
        if location == "left":
            tanslate = _bboxhist.width + space + axspace
            _bboxhist = _bboxhist.translated(bboxorig.xmin-_bboxhist.xmin+space, 0)
            _bboxax = _bboxax.translated(tanslate,0)
        
    else:
        raise ValueError("location must be 'top'/'bottom'/'left' or 'right'")


    axhist.set_position(_bboxhist)
    ax.set_position(_bboxax)

    # ---------------------
    # remove their ticks
    if shareax:
        if location in ["top","right"]:
            [[label.set_visible(False) for label in lticks]
            for lticks in [axhist.get_xticklabels(),axhist.get_yticklabels()]]
        elif location == "bottom":
            [[label.set_visible(False) for label in lticks]
            for lticks in [ax.get_xticklabels(),axhist.get_yticklabels()]]
        elif location == "left":
            [[label.set_visible(False) for label in lticks]
            for lticks in [ax.get_yticklabels(),axhist.get_xticklabels()]]
    
    return axhist



##############################
#                            #
#  Convertion Tools          #
#                            #
##############################
# ---------------------- #
#  Sphere to Cartesian   #
# ---------------------- #
def sph2cart(vec):
    """
    Convert vector in spherical coordinates (r, theta, phi ; angles in degrees)
    to Cartesian coordinates [x,y,z].
    
    Returns
    -------
    x,y,z
    """
    v, l, b = vec[0], vec[1]*_DEG2RA, vec[2]*_DEG2RA
    return np.asarray([v*np.cos(b)*np.cos(l), 
                       v*np.cos(b)*np.sin(l), 
                       v*np.sin(b)])    

def cart2sph(vec):
    """
    Convert vector in Cartesian coordinates [x,y,z] to spherical coordinates [r, theta, phi]
    (angles in degrees).
    """
    x, y ,z = vec
    v = np.sqrt(x**2 + y**2 + z**2)
    return np.array([v,
                    (np.arctan2(y,x) / _DEG2RA + 180) % 360 - 180, 
                     np.arcsin(z/v) / _DEG2RA])
     
# ---------------------- #
#  Rotation              #
# ---------------------- #

def rot_xz(v, theta):
    """
    Rotate Cartesian vector v [x,y,z] by angle theta around axis (0,1,0)
    """
    return np.asarray([v[0]*np.cos(theta*_DEG2RA) - v[2]*np.sin(theta*_DEG2RA),
                      v[1],
                      v[2]*np.cos(theta*_DEG2RA) + v[0]*np.sin(theta*_DEG2RA)])

def rot_xz_sph(l, b, theta):
    """
    Rotate Spherical coordinate (l,b = theta, phi) by angle theta around axis (0,1,0)
    """
    v_rot = rot_xz(sph2cart([1,l,b]), theta)
    return cart2sph(v_rot)[1:]





#!/usr/bin/env python
#

import os, hashlib
import sys
import requests
import warnings
import numpy as np
LOGIN_URL = "https://irsa.ipac.caltech.edu/account/signon/login.do"

import base64
if sys.version_info > (3,0):
    from configparser import ConfigParser
    _PYTHON3 = True
else:
    from ConfigParser import ConfigParser
    _PYTHON3 = False
    
_SOURCEDIR = os.path.dirname(os.path.realpath(__file__))


_ENCRYPT_FILE = os.path.expanduser("~")+"/.ztfquery"
_ENCRYPTING_FILE = os.path.expanduser("~")+"/.queryirsa"


LOCALSOURCE   = os.getenv('ZTFDATA',"./Data/")


# ================= #
#  Crypting         #
# ================= #

def _load_id_(which, askit=True):
    """ returns login information for the requested enty"""
    import base64
    config = ConfigParser()
    config.read( _ENCRYPT_FILE )
    if which not in config.sections():
        if not askit:
            raise AttributeError(f"No {which} account setup. Add then in .ztfquery or run ztfquery.io.set_account({which})"%(which,which))
        else:
            warnings.warn(f"No {which} account setup, please provide it")
            set_account(which)
            config = ConfigParser()
            config.read( _ENCRYPT_FILE )

    if _PYTHON3:
        return config[which.lower()]["username"], base64.b64decode(config[which.lower()]["password"][2:-1]).decode("utf-8")
    else:
        return config.get(which.lower(),"username"), base64.b64decode(config.get(which.lower(),"password"))

def set_account(which, username=None, password=None, test=True, force=False):
    """ Setup the username and password (simply encrypted!) for the given `which` account. 
    Saved in ~/.ztfquery
    """
    import base64
    import getpass
    config = ConfigParser()
    config.read( _ENCRYPT_FILE )
    # - Name & Password
    if username is None:
        if _PYTHON3:
            username = input('Enter your %s login: '%which)
        else:
            username = raw_input('Enter your %s login: '%which)
            
    if password is None:
        password = getpass.getpass()
        
    # - Check inputs
    if test:
        wrong_ = False
        if which == "irsa":
            if not test_irsa_account([username, password]):
                warnings.warn("The irsa_test for you account returns False. Most likely you provided incorrect logins")
                wrong_ = True
        else:
            warnings.warn("No test designed for %s. Cannot test if logins are correct."%which)
        if wrong_ and not force:
            raise ValueError("Bad username/passworg for %s. force=False so the logins are not stored. "%which)
            
    password_ = base64.b64encode( password.encode("utf-8") )
    if _PYTHON3:
        config[which.lower()] = {"username":username, "password": password_ }        
    else:
        section = which.lower()
        if section not in config.sections():
            config.add_section(section)
        config.set(section,"username",username)
        config.set(section,"password",password_)

        
    with open( _ENCRYPT_FILE , 'w') as configfile:
        config.write(configfile)
        

#
# TEST
#
# - Password testing
def test_irsa_account(auth=None):
    """  returns True if the IRSA account is correctly set. """
    if auth is None:
        auth = _load_id_("irsa")
    return ".ipac.caltech.edu" in get_cookie(*auth)._cookies



# - File testing
def get_localfiles(extension="*", startpath=None):
    """ Look for all file with the given extension recursively starting from `startpath`.
    (based on glob)
    
    Parameters
    ----------
    extension: [string] -optional-
        All the 'file.{}'.format(extension) will be looked at.
        (first '.' ignored such that extension='.fits' is equivalent to extension='fits')

    startpath: [None or path] -optional-
        From which directory does this start to look at.
        If None: $ZTFDATA (stored as io.LOCALSOURCE) will be used.

    Returns
    -------
    list of file.
    """
    from glob import glob
    if startpath is None:
        startpath = LOCALSOURCE
    if extension.startswith("."):
        extension = extension[1:]
        
    return [f for f in glob(startpath + "**/*.%s"%extension, recursive=True)]

def run_full_filecheck(extension="*", startpath=None,
                        erasebad=True, redownload=False, 
                        nprocess=4, show_progress=True, notebook=False,
                        **kwargs ):
    """ Look for all file with the given extension recursively starting from `startpath` and checks if the file 
    is usable ok not. This returns the bad files.
    
    Parameters
    ----------

    // Data files
    extension: [string] -optional-
        All the 'file.{}'.format(extension) will be looked at.
        (first '.' ignored such that extension='.fits' is equivalent to extension='fits')

    startpath: [None or path] -optional-
        From which directory does this start to look at.
        If None: $ZTFDATA (stored as io.LOCALSOURCE) will be used.

    // Check options

    erasebad: [bool] -optional-
        Do you want to remove from your local directory the corrupted files ?
        
    redownload: [bool] -optional-
        Shall corrupted file be automatically re downloaded ?
        (Only works for IRSA files ('/sci/','/raw/', '/ref/', '/cal/')

    nprocess: [int] -optional-
        Number of paralell processing

    show_progress: [bool] -optional-
        Do you want to show the progress bar ?
        
    notebook: [bool]
        Are you running from a notebook. 
        Ignored if show_progress=False

    Returns
    -------
    list of corrupted/bad files (might already be removed, see erasebad)

    """
    all_ztffiles = get_localfiles(extension=extension, startpath=startpath)
    print("%d files to check"%len(all_ztffiles))
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        badfiles = test_files(all_ztffiles, erasebad=erasebad, nprocess=nprocess,
                             show_progress=show_progress, notebook=notebook,
                             redownload=redownload, **kwargs)
        
    return badfiles
    
def test_files(filename, erasebad=True, nprocess=1, show_progress=True, notebook=False,
                   redownload=False, **kwargs ):
    """ 
    
    Parameters
    ----------
    filename: [fiulepath or list of]
        File(s) to be checked. 

    erasebad: [bool] -optional-
        Do you want to remove from your local directory the corrupted files ?
        
    redownload: [bool] -optional-
        Shall corrupted file be automatically re downloaded ?
        (Only works for IRSA files ('/sci/','/raw/', '/ref/', '/cal/')

    nprocess: [int] -optional-
        Number of paralell processing

    show_progress: [bool] -optional-
        Do you want to show the progress bar ?
        
    notebook: [bool]
        Are you running from a notebook. 
        Ignored if show_progress=False
       
    Returns
    -------
    list of corrupted/bad files (might already be removed, see erasebad)

    """
    if nprocess is None:
        nprocess = 1
    elif nprocess<1:
        raise ValueError("nprocess must 1 or higher (None means 1)")

    filename = np.atleast_1d(filename)
    
    if nprocess == 1:
        fileissue = [f for f in filename if not _test_file_(f, erasebad=erasebad,
                                                            redownload=redownload,
                                                            **kwargs)]
    else:
        import multiprocessing
        if show_progress:
            from astropy.utils.console import ProgressBar
            bar = ProgressBar( len(filename), ipython_widget=notebook)
        else:
            bar = None

        erasebad_   = [erasebad]*len(filename)
        fileissue = []
        
        with multiprocessing.Pool(nprocess) as p:
            # Da Loop
            for j, isgood in enumerate( p.imap(_test_file_multiprocess_, zip(filename, erasebad_)) ):
                if bar is not None:
                    bar.update(j)
                if not isgood:
                    fileissue.append(filename[j])
                    
            if bar is not None:
                bar.update( len(filename) )
                
    if len(fileissue) >0:
        warnings.warn("%d file failed"%len(fileissue))
        if redownload:
            from .buildurl import _localsource_to_source_
            to_download_urls, locations = np.asarray([_localsource_to_source_(filename)
                                                        for filename in fileissue]).T
            source_to_dl = ["irsa"]
            for source in source_to_dl:
                source_dl = np.in1d(locations, [source])
                print("Downloading %d files from %s"%(len(source_dl[source_dl]), source))
                download_url(np.asarray(to_download_urls)[source_dl], np.asarray(fileissue)[source_dl],
                                 show_progress=show_progress, notebook=notebook, verbose=True,
                                 overwrite=True, nprocess=nprocess, cookies=get_cookie(*_load_id_(source)),
                         **kwargs)
            for source_ in np.unique(locations):
                if source_ is not None and source_ not in source_to_dl:
                    warnings.warn("files from %s have not downloaded (not implemented)."%source_)
                
        return fileissue
        

def _test_file_multiprocess_(args):
    """ """
    filename, erasebad = args
    return _test_file_(filename, erasebad=erasebad, fromdl=False, redownload=False)

def _test_file_(filename, erasebad=True, fromdl=False,
                redownload=False, verbose=True):
    """ """
    propissue = dict(erasebad=erasebad, fromdl=fromdl, redownload=redownload, verbose=verbose)
    # Fits file
    if ".fits" in filename:
        from astropy.io import fits
        if not hash_for_file_exists(filename):
            try:
                _ = fits.getdata(filename)
                calculate_and_write_hash(filename)
            except FileNotFoundError:
                warnings.warn("[Errno 2] No such file or directory: %s"%filename)
            except:
                _fileissue_(filename, **propissue)
                return False
        
    # txt file        
    elif ".txt" in filename:
        if not hash_for_file_exists(filename):
            try:
                _ = open(filename).read().splitlines()
                calculate_and_write_hash(filename)
            except FileNotFoundError:
                warnings.warn("[Errno 2] No such file or directory: %s"%filename)
            except:
                _fileissue_(filename, **propissue)
                return False
        
    # other extensions        
    else:
        warnings.warn("no file testing made for .%s files"%filename.split(".")[-1])
            
    return True

def _fileissue_(filename, erasebad=True, fromdl=False, redownload=False, verbose=True):
    """ """
    if fromdl:
        warnings.warn("Download failed %s seems corrupted (cannot open)"%filename)
    else:
        warnings.warn("cannot open file %s"%filename)
        
    if erasebad:
        warnings.warn("removing %s")
        os.remove(filename)
    else:
        warnings.warn("%s NOT ERASED")

    if redownload:
        from .buildurl import _localsource_to_source_
        url_to_dl, location = _localsource_to_source_(filename)
        if url_to_dl is not None:
            download_single_url(url_to_dl, fileout=filename, overwrite=True, verbose=verbose,
                                cookies = get_cookie(*_load_id_(location))
                                )
        else:
            warnings.warn("No url to donwload, redownload ignored")
    
# ================= #
#   Logging Tools   #
# ================= #
def get_cookie(username, password):
    """Get a cookie from the IPAC login service

    Parameters
    ----------
    username: [str]
        The IPAC account username
    password: [str]
        The IPAC account password
    """
    url = "%s?josso_cmd=login&josso_username=%s&josso_password=%s" % (LOGIN_URL, username, password)
    return requests.get(url).cookies


def _download_(args):
    """ To be used within _ZTFDownloader_.download_data() 
    url, fileout,overwrite,verbose = args
    """
    url, fileout,  overwrite, verbose = args
    download_single_url(url, fileout=fileout, overwrite=overwrite, verbose=verbose)


def download_from_filename(filename, suffix=None, verbose=False, overwrite=False,
                               auth=None, **kwargs):
    """ Download the file associated to the given filename """
    from .buildurl import filename_to_scienceurl
    if auth is None:
        auth = _load_id_("irsa")
    cookies = get_cookie(*auth)
        
    irsa_filename = filename_to_scienceurl(filename, suffix=suffix, verbose=verbose, source="irsa")
    local_filename = filename_to_scienceurl(filename, suffix=suffix, verbose=verbose, source="local")
    
    download_url(np.atleast_1d(irsa_filename),
                 np.atleast_1d(local_filename),
                 overwrite=overwrite,
                 cookies = cookies, **kwargs)
    
    
    
def download_url(to_download_urls, download_location,
                show_progress = True, notebook=False, verbose=True,
                overwrite=False, nprocess=None, cookies=None,
                **kwargs):
    """ """
    if nprocess is None:
        nprocess = 1
    elif nprocess<1:
        raise ValueError("nprocess must 1 or higher (None means 1)")

    if nprocess == 1:
        # Single processing
        if verbose:
            warnings.warn("No parallel downloading")
        for url, fileout in zip(to_download_urls, download_location):
            download_single_url(url,fileout=fileout, show_progress=show_progress,
                                    notebook=notebook, 
                                    overwrite=overwrite, verbose=verbose, cookies=cookies, **kwargs)
    else:
        # Multi processing
        import multiprocessing
        if show_progress:
            from astropy.utils.console import ProgressBar
            bar = ProgressBar( len(to_download_urls), ipython_widget=notebook)
        else:
            bar = None
                
        if verbose:
            warnings.warn("parallel downloading ; asking for %d processes"%nprocess)

        # Passing arguments
        overwrite_ = [overwrite]*len(to_download_urls)
        verbose_   = [verbose]*len(to_download_urls)
        with multiprocessing.Pool(nprocess) as p:
            # Da Loop
            for j, result in enumerate( p.imap_unordered(_download_, zip(to_download_urls,
                                                                    download_location,
                                                                 overwrite_, verbose_))):
                if bar is not None:
                    bar.update(j)
                    
            if bar is not None:
                bar.update( len(to_download_urls) )

            
def download_single_url(url, fileout=None, 
                        overwrite=False, verbose=True, cookies=None,
                        show_progress=True, notebook=False, chunk=1024,
                        filecheck=True, erasebad=True,
                        **kwargs):
    """ Download the url target using requests.get.
    the data is returned (if fileout is None) or stored in `fileout`
    Pa
    """
    if fileout is not None and not overwrite and os.path.isfile( fileout ):
        if verbose:
            warnings.warn("%s already exists: skipped"%fileout)
        return
    else:
        if verbose and fileout:
            warnings.warn("downloading %s to %s"%(url,fileout))

    # = Password and Username
    if cookies is None:
        cookies = get_cookie(*_load_id_("irsa"))

    # - requests options 
    download_prop = dict(cookies=cookies, stream=True)
    for k,v in kwargs.items():
        download_prop[k] = v
        
    if cookies in ["no_cookies"]:
        _ = download_prop.pop("cookies")

    request_fnc = "get" if not "data" in download_prop else "post"
    # = Where should the data be saved?
    if fileout is not None:
        directory = os.path.dirname(fileout)
        if not os.path.exists(directory):
            os.makedirs(directory)

    else:
        download_prop["stream"] = False
        return getattr(requests,request_fnc)(url, **download_prop)

    # With Progress bar?
    if not show_progress:
        response = getattr(requests,request_fnc)(url, **download_prop)
        if response.status_code == 200:
            with open(fileout, 'wb') as f:
                for data in response.iter_content(chunk):
                    f.write(data)
    
    else:
        from astropy.utils.console import ProgressBar
        response = getattr(requests,request_fnc)(url, **download_prop)
        if response.status_code == 200:
            chunk_barstep = 500
            f = open(fileout, 'wb')
            with ProgressBar(int(response.headers.get('content-length'))/(chunk_barstep*chunk),
                             ipython_widget=notebook) as bar:
                for i,data in enumerate(response.iter_content(chunk_size=chunk)):
                    if i%chunk_barstep==0:
                        bar.update()
                    f.write(data)
            f.close()
            calculate_and_write_hash(fileout)

    if filecheck:
        _test_file_(fileout, erasebad=erasebad, fromdl=True)

def calculate_hash(fname):
    """ """
    f = open(fname, 'rb')
    hash_md5 = hashlib.md5()
    for chunk in iter(lambda: f.read(4096), b""):
        hash_md5.update(chunk)
    hexdigest = hash_md5.hexdigest()
    f.close()
    return hexdigest

def calculate_and_write_hash(fname):
    """ """
    f = open(fname, 'rb')
    hash_md5 = hashlib.md5()
    for chunk in iter(lambda: f.read(4096), b""):
        hash_md5.update(chunk)
    hexdigest = hash_md5.hexdigest()
    f.close()
    f_hash = open(f"{fname}.md5", 'w')
    f_hash.write(hexdigest)
    f_hash.close()

def read_hash(fname):
    """ """
    f_hash = open(f"{fname}.md5", 'r')
    hash_md5 = f_hash.read()
    return hash_md5

def compare_hash(fname):
    """ """
    f_hash = open(hash_fname, 'r')
    hash_md5_read = f_hash.read()
    hash_md5_calculated = calculate_hash(fname)
    if hash_md5_read == hash_md5_calculated:
        return True
    else:
        return False

def hash_for_file_exists(fname):
    """ """
    return os.path.exists(f"{fname}.md5")



#! /usr/bin/env python
# -*- coding: utf-8 -*-

""" ZTF monitoring information """

import requests
import os, json
import warnings
import numpy as np
import pandas

from astropy import time

from io import StringIO
from . import io, fields, ztftable

SKYVISIONSOURCE = os.path.join(io.LOCALSOURCE,"skyvision")
if not os.path.exists( SKYVISIONSOURCE ):
    os.mkdir(SKYVISIONSOURCE)

#############################
#                           #
# Stand Alone Functions     # 
#                           #
#############################
def get_log(date, which="completed", download=True, update=False, **kwargs):
    """ Get target spectra from the marshal. 
    
    Parameters
    ----------
    date: [string]
        Date of the log with the format YYYY-MM-DD

    download: [bool] -optional-
        Should the spectra be downloaded if necessary ?

    update: [bool] -optional-
        Force the re-download of the completed_log.

    Returns
    -------
    DataFrame
    """
    if len(np.atleast_1d(date))>1:
        return pandas.concat([get_log(date_, which=which, download=download, update=update, **kwargs)  for date_ in date])
    
    date = np.atleast_1d(date)[0]
    if update:
        download_log(date, which=which, store=True, **kwargs)
    
    logdf = get_local_log(date, which=which, safeout=True)
    if logdf is None:
        if update:
            warnings.warn(f"Download did not seem successful. Cannot retreive the {which}_log for {date}")
            return None
        elif not download:
            warnings.warn(f"No local {which}_log for {date}. download it or set download to true")
            return None
        
        return get_log(date, which=which, update=True, **kwargs)
    else:
        return logdf

# -------------- #
#  Detailed      #
# -------------- #
def get_log_filepath(date, which="completed"):
    """ local filepath of the completed_log for the given date """
    return os.path.join(SKYVISIONSOURCE, f"{date}_{which}_log.csv")

def get_local_log(date, which="completed", safeout=False):
    """ """
    filein = get_log_filepath(date, which=which)
    if not os.path.isfile(filein):
        if safeout:
            return None
        raise IOError(f"No {which}_log locally stored for {date}. see download_log()")
    
    return pandas.read_csv(filein)

def get_daterange(start, end=None):
    """ """
    if end is None:
        import datetime
        yesterday = datetime.date.today() - datetime.timedelta(days=1)
        end = yesterday.isoformat()

    # Dates to be downloaded.
    return [f"{i.year}-{i.month:02d}-{i.day:02d}" for i in pandas.Series(pandas.date_range(start,end, freq='D'))]

def download_log(date, which="completed", auth=None, store=True, **kwargs):
    """ Generic downloading function for the logs. 
    Calls the individual download_{which}_log
    """
    return eval(f"download_{which}_log")(date, auth=auth, store=store, **kwargs)

def download_timerange_log(start="2018-03-01", end=None, which="completed",
                            nprocess=1, auth=None,
                            show_progress=True, notebook=True, verbose=True):
    """ Storing and not return forced. See download_completed_log() for individual date downloading. """
    if nprocess is None:
        nprocess = 1
    elif nprocess<1:
        raise ValueError("nprocess must 1 or higher (None means 1)")

    if auth is not None:
        print("updating skyvision authentification")
        io.set_account("skyvision", username=auth[0], password=auth[0], test=False)
    if which not in ["completed", "qa"]:
        raise ValueError(f"which can only be completed of qa, {which} given")
    
    dl_function = eval(f"download_{which}_log")
    
    # Dates to be downloaded.
    dates = get_daterange(start, end=None)
    
    if nprocess == 1:
        # Single processing
        if verbose:
            warnings.warn("No parallel downloading")
        
        _ = [dl_function(date, auth=auth, store=True, returns=False) for date in dates]
    else:
        # First download manually otherwise it could fail, unclear why
        _ = dl_function(dates[0], auth=auth, store=True, returns=False)
        # Multi processing
        import multiprocessing
        if show_progress:
            from astropy.utils.console import ProgressBar
            bar = ProgressBar( len(dates[1:]), ipython_widget=notebook)
        else:
            bar = None
                
        if verbose:
            warnings.warn(f"parallel downloading ; asking for {nprocess} processes")

        # Passing arguments
        with multiprocessing.Pool(nprocess) as p:
            # Da Loop
            for j, result in enumerate( p.imap(dl_function, dates[1:])):
                if bar is not None:
                    bar.update(j)
                    
            if bar is not None:
                bar.update( len(dates[1:]) )

# ================= #
#  LOG Downloading  #
# ================= #
def download_completed_log(date, auth=None, store=True,
                            returns=True, set_columns=True, verbose=False):
    """ 
    Parameters
    ----------
    date: [string]
        Date with the following format: YYYY-MM-DD

    auth: [2d-list or None] -optional-
        Skyvision.caltech.edu/ztf login and password.

    
    Returns
    -------
    DataFrame
    """
    if date in ["2018-03-04"]:
        warnings.warn(f"Known night with Failure only {date}, log format incorrect")
        logtable=None
    else:
        response = requests.post(f"http://skyvision.caltech.edu/ztf/queue/?obsdate={date}",
                                auth = io._load_id_("skyvision") if auth is None else auth)
        logtable = response.text
        
    if verbose:
        print(logtable)
    if logtable is None:
        warnings.warn(f"no downloaded information for night {date}")
        data = None
    elif "does not exist for this night" in logtable:
        warnings.warn(f"No observing log for {date}")
        data = None
    else:
        data = [l.split() for l in logtable.splitlines()] if logtable is not None else None
            
    if verbose:
        print(data)
    if time.Time(date)<=time.Time("2018-07-09"):
        columns = ['UT Date','UT Time','Sequence ID', 'Program ID', 'Field ID', 'RA', 'DEC', 'Epoch',
                   'RA Rate', 'Dec Rate', 'Exptime', 'Filter', 'Observation Status', 'Setup Time', 'Exptime']
    else:
        columns = ['UT Date','UT Time', 'Base Image Name','Sequence ID', 'Program ID', 'Field ID', 'RA', 'DEC', 'Epoch',
                   'RA Rate', 'Dec Rate', 'Exptime', 'Filter', 'Observation Status', 'Setup Time', 'Exptime']

            
    try:
        df = pandas.DataFrame(data, columns=columns)
    except:
        warnings.warn(f"Column format does not match the completed_log date downloade for {date}")
        print(data)
        return None
        
    if store:
        df.to_csv( get_log_filepath(date, which="completed"), index=False)
    if returns:
        return df

def download_qa_log(date, auth=None, summary_values=None, store=True,
                        where_statement="", groupby_values=False, returns=True):
    """ 
    Parameters
    ----------
    where_statement: [string]
        additional SQL query, e.g.:
        ' AND programid in (1,2)' for limiting yourself to MSIP+Parners

    """
    if summary_values is None:
        summary_values = ['obsdatetime', 'nightdate','obsjd',
                          'exptime', 'ccdid','qid', "rcid","fid",
                          'scibckgnd','sciinpseeing','scisat','nsexcat',
                          'refbckgnd','refinpseeing','refsat',
                          'programid','maglimit', 'field', 'fwhm','status','statusdif', 
                          'qcomment']
            
    url = "http://skyvision.caltech.edu/ztf/status/summary_table"
    payload = {'summary_values': summary_values,
               'obsdate': date,
               'where_statement': where_statement,
               'groupby_values': groupby_values,
               'return_type': "csv",
               'time_between': False,
               'add_summary_map': False,
               }

    headers = {'content-type': 'application/json'}
    json_data = json.dumps(payload)
    response = requests.post(url,
                             data=json_data,
                             auth= io._load_id_("skyvision") if auth is None else auth,
                             headers=headers)
    
    
    y = json.loads(response.text)
    
    df = pandas.read_csv(StringIO(y['obstable']), header=0,
                     low_memory=False, index_col=0)

    df['obsdatetime'] = pandas.to_datetime(df['obsdatetime'])
    
    if store:
        df.to_csv( get_log_filepath(date, which="qa"), index=False)
        
    if returns:
        return df

# ================ #
#                  #
#   LOG Classes    #
#                  #
# ================ #

class ZTFLog( ztftable._ZTFTable_ ):
    """ """
    _NAME_ = "ztflog"
    
    def __init__(self, dataframe ):
        """ Date or list of date """
        if dataframe is not None:
            self.set_logs( dataframe)

    @classmethod
    def from_date(cls, date, load_data=True, load_obsjd=True,
                      download=True, update=False, **kwargs):
        """ Logs a date or a list of dates"""
        logdf = get_log(np.atleast_1d(date), which=cls._NAME_, download=download, update=update, **kwargs)
        return cls(  logdf )

        
    @classmethod
    def from_daterange(cls, start="2018-03-01", end=None, load_data=True, load_obsjd=True,
                           download=True, update=False, **kwargs):
        """ """
        if start is None:
            start = "2018-03-01"

        dates = get_daterange(start=start, end=end)
        logdf = get_log(np.atleast_1d(dates), which=cls._NAME_, download=download, update=update, **kwargs)
        return cls( logdf )

     # =============== #
    #   Methods       #
    # =============== #
    # -------- #
    #  LOADER  #
    # -------- #
    def load_data(self):
        """ """
        self._data = self.logs
        
    # -------- #
    #  SETTER  #
    # -------- #    
    def set_logs(self, logs):
        """ """
        self._logs = logs

    # -------- #
    #  GETTER  #
    # -------- #
    
    # -------- #
    # PLOTTER  #
    # -------- #
    # =============== #
    #  Properties     #
    # =============== #
    @property
    def logs(self):
        """ original logs as given by skyvision """
        return self._logs
    
    @property
    def data(self):
        """ clean (renamed and reshaped) logs """
        if not hasattr(self,"_data"):
            self.load_data()
        return self._data
    
    def has_multiple_logs(self):
        """ """
        return type(self.logs.index) is pandas.MultiIndex
    
# ============== #
#                #
#  COMPLETED     #
#                #
# ============== #
class CompletedLog( ZTFLog ):
    """ """
    _NAME_ = "completed"
    # =============== #
    #   Methods       #
    # =============== #
    # -------- #
    #  LOADER  #
    # -------- #        
    def load_data(self, load_obsjd=False):
        """ """
        lm = self.get_completed_logs()
        dict_= {"datetime": np.asarray(lm["UT Date"] +"T"+ lm["UT Time"], dtype=str),
                "date":lm["UT Date"],
                "exptime":lm["Exptime"].astype(float),
                "totalexptime":lm["Setup Time"].astype(float),
                "fid":lm["Filter"].apply(lambda x: 1 if x=="FILTER_ZTF_G" else 2 if x=="FILTER_ZTF_R" else 3),
                "field":lm["Field ID"].astype(int),
                "pid":lm["Program ID"], # 0: inge ; 1: MSIP ; 2: Parners ; 3: Caltech
                "ra": lm["RA"],
                "dec": lm["DEC"],
                "totaltime":lm["Setup Time"]
                }
        self._data = pandas.DataFrame(dict_)
        if load_obsjd:
            self.load_obsjd()

    def load_obsjd(self):
        """ converts the datetime format in obsjd """
        self.data.loc[:, "obsjd"] = time.Time(np.asarray(self.data.datetime, dtype="str")).jd
        
    def get_obsjd(self, update=False):
        """ get the data.datetime in obsjd format. Loads it the first time you call it."""
        if "obsjd" not in self.data.columns:
            self.load_obsjd()
        return self.data.obsjd
    
    # -------- #
    #  GETTER  #
    # -------- #
    def get_when_target_observed(self, radec, pid=[1,2,3], fid=None, startdate=None, enddate=None,
                                    query=None, **kwargs):
        """ Returns the data raws corresponding to the given coordinates
    
        Parameters
        ----------
        radec: [float,float]
            target coordinates in degree
        
        pid: [int or list of] -optional-
            Selection only cases observed as part as the given program id.s
            (0: engineering ; 1: MSIP ; 2:Parners ; 3:CalTech )
            None means no selection.
        
        fid: [int or list of] -optional-
            Selection only cases observed with the given filters
            (1: ztf:g ; 2: ztf:r ; 3: ztf:i)
            None means no selection.
        
        startdate, enddate: [string] -optional-
            Select the time range to be considered. None means no limit.
        
        Returns
        -------
        DataFrame
        """
        target_fields = fields.get_fields_containing_target(*radec)
        return self.get_when_field_observed( target_fields, pid=pid, fid=fid,
                                             startdate=startdate, enddate=enddate, query=query, **kwargs)
    
    def get_when_field_observed(self, field, pid=[1,2,3],
                                    fid=None, startdate=None, enddate=None, query=None, **kwargs):
        """ Returns the data raws corresponding to the given filters
    
        Parameters
        ----------
        fields: [int or list of]
            Field(s) id you want. If multiple field returns an 'or' selection applies.
        
        pid: [int or list of] -optional-
            Selection only cases observed as part as the given program id.s
            (0: engineering ; 1: MSIP ; 2:Parners ; 3:CalTech )
            None means no selection.
        
        fid: [int or list of] -optional-
            Selection only cases observed with the given filters
            (1: ztf:g ; 2: ztf:r ; 3: ztf:i)
            None means no selection.
        
        startdate, enddate: [string] -optional-
            Select the time range to be considered. None means no limit.
        
        Returns
        -------
        DataFrame
        """
        
        return self.get_filtered(field=field, pid=pid, fid=fid, startdate=startdate, enddate=enddate, query=query, **kwargs)

    def get_filtered(self, field=None, fid=None, pid=None, startdate=None, enddate=None, grid=None, query=None):
        """  
        Parameters
        ----------
        query: [string] -optional-
            Generic pandas.DataFrame query regex call, i.e. self.data.query(query)

        fields: [int or list of]
            Field(s) id you want. If multiple field returns an 'or' selection applies.
        
        pid: [int or list of] -optional-
            Selection only cases observed as part as the given program id.s
            (0: engineering ; 1: MSIP ; 2:Parners ; 3:CalTech )
            None means no selection.
        
        fid: [int or list of] -optional-
            Selection only cases observed with the given filters
            (1: ztf:g ; 2: ztf:r ; 3: ztf:i)
            None means no selection.
        
        startdate, enddate: [string] -optional-
            Select the time range to be considered. None means no limit.
        
        Returns
        -------
        pandas.Serie of bool
        """
        queried = super().get_filtered(field=field, fid=fid, grid=grid, query=query)
        if pid is None and (startdate is None and enddate is None):
            return queried
        
        pidflag = True if pid is None else queried["pid"].isin(np.atleast_1d(pid))
        # DateRange Selection
        if startdate is None and enddate is None:
            dateflag = True
        elif startdate is None:
            dateflag = queried["date"]<=enddate
        elif enddate is None:
            dateflag = queried["date"]>=startdate
        else:
            dateflag = queried["date"].between(startdate,enddate)

        return queried[pidflag & dateflag]
            
    def get_date(self, date):
        """ """
        return self.data[self.data["date"].isin(np.atleast_1d(date))]

    def get_loaded_dates(self, which="data"):
        """ """
        return np.unique(self.data["date"])
        
    def get_completed_logs(self):
        """ """
        return self.logs[self.logs["Observation Status"]=="COMPLETE"]

    # -------- #
    # PLOTTER  #
    # -------- #
    def show_survey(self, grid="main", show_mw=True, interval=5):
        """ """
        from ztfquery import fields
        
        field_id,filterid,dates = self.data[ ["field","fid","datetime"] ].values.T
        fid_color = [fields.FIELD_COLOR[i] for i in filterid]
        
        fanim = fields.FieldAnimation(field_id, dates=dates, facecolors=fid_color)
        if grid is not None:
            fanim.show_ztf_grid(grid)
        if show_mw:
            fanim.show_milkyway()
        fanim.launch(interval=interval)
        return fanim
    
# ============== #
#                #
#  QA            #
#                #
# ============== #
class QALog( ZTFLog ):
    """ """
    _NAME_ = "qa"
    def set_logs(self, logs):
        """ """
        self._logs  = logs.rename(columns={"programid":"pid"})

    def load_data(self, pid=[1,2,3]):
        """ """
        self._data = self.logs.query("pid in @pid")



#! /usr/bin/env python
#

""" Combine MetaSearch and MetaURL to get data from IRSA """
import os
import numpy as np
from .metasearch import download_metadata, _test_kind_
from . import buildurl
import warnings


#############################
#                           #
#   Main Query Tools        #
#                           #
#############################
class _ZTFTableHandler_( object ):
    """ """
    # -------------- #
    #  FIELDS        #
    # -------------- #
    def get_observed_fields(self, grid="both"):
        """ get the (unique) list of field observed  type. """
        if "field" not in self._data.columns:
            return None
        all_fields = np.unique(self._data["field"])
        if grid is None or grid in ["both"]:
            return all_fields
        
        from .fields import fields_in_main
        if grid in ["main", "first", "primary"]:
            return all_fields[fields_in_main(all_fields)]
        elif grid in ["other", "secondary"]:
            return all_fields[~fields_in_main(all_fields)]
        else:
            raise ValueError("Cannot parse the given grid %s"%grid)
        
    def get_field_average_value(self, value, grid="both", fid=[1,2,3]):
        """ """
        flagfield = True if fid is None or "fid" not in self._data.columns else np.in1d(np.asarray(self._data["fid"], dtype="int"), fid)
        return {f_: np.nanmean(self._data[np.in1d(self._data["field"], f_) * flagfield][value])
                    for f_ in self.get_observed_fields(grid=grid)}
        
    def get_field_obsdensity(self, grid="both", fid=[1,2,3]):
        """ """
        flagfield = True if fid is None or "fid" not in self._data.columns else np.in1d(np.asarray(self._data["fid"], dtype="int"), fid)
        return {f_: len(self._data[np.in1d(self._data["field"], f_) * flagfield]) for f_ in self.get_observed_fields(grid=grid)}

    def show_fields(self, field_val,
                    ax=None,
                    show_ztf_fields=True,
                    colorbar=True, cax=None, clabel=" ", 
                    cmap="viridis",origin=180,
                    vmin=None, vmax=None,  **kwargs):
        """ 
        Parameters
        ----------
        colored_by: 
        """
        import matplotlib.pyplot as mpl
        from .fields import display_field
        if "field" not in self._data.columns:
            raise AttributeError("No 'field' entry available")
        if origin != 180:
            warnings.warn("Only the origin 180 has been implemented. boundaries issue arises otherwise. origin set to 180")
            origin = 180
            
        tick_labels = np.array([150, 120, 90, 60, 30, 0, 330, 300, 270, 240, 210])
        tick_labels = np.remainder(tick_labels+360+origin,360)

        # - Axes definition
        if ax is None:
            fig = mpl.figure(figsize=(8,5))
            ax = fig.add_subplot(111, projection="hammer")
        else:
            fig = ax.figure

        ax.set_xticklabels(tick_labels)     # we add the scale on the x axis
        # - Plotting
        if show_ztf_fields:
            from .fields import show_ZTF_fields
            show_ZTF_fields(ax)

        # Removing the NaNs
        field_val = {f:v for f,v in field_val.items() if not np.isnan(v)}
        values = list(field_val.values())
        if len(values)==0 or not np.any(values):
            if cax is not None:
                cax.set_visible(False)
            return
        
        if vmin is None: vmin = "0"
        if type(vmin) == str: vmin=np.percentile(values, float(vmin))
        if vmax is None: vmax = "100"
        if type(vmax) == str: vmax=np.percentile(values, float(vmax))
        if type(cmap) == str: cmap = mpl.get_cmap(cmap)
        for f,v in field_val.items():
            display_field(ax, f, facecolor=cmap((v-vmin)/(vmax-vmin)) if vmax-vmin !=0 else cmap(0),origin=origin, 
                              **kwargs)



        if colorbar:
            if vmax-vmin !=0:
                from .utils.tools import insert_ax, colorbar
                if cax is None: cax = insert_ax(ax, "bottom",
                                            shrunk=0.93, space=-0.0, axspace=0.02)
                colorbar(cax, cmap, vmin=vmin, vmax=vmax, label=clabel)
            elif cax is not None:
                cax.set_visible(False)
        
        
    def show_gri_fields(self, title=" ",
                        show_ztf_fields=True,
                        colorbar=True, 
                        colored_by="visits", grid="main",
                        **kwargs):
        """  """
        import matplotlib.pyplot as mpl
        from .fields import FIELDS_COLOR
        fig = mpl.figure(figsize=[9,6])
        fig.suptitle(title, fontsize="large")
        # G
        axg   = fig.add_axes([0.03,0.52,0.43,0.48], projection="hammer")
        caxg  = fig.add_axes([0.03,0.54,0.43,0.015])
        axg.tick_params(labelsize="x-small", labelcolor="0.3" )
        # R
        axr   = fig.add_axes([0.54,0.52,0.43,0.48], projection="hammer")
        caxr  = fig.add_axes([0.54,0.54,0.43,0.015])
        axr.tick_params(labelsize="x-small", labelcolor="0.3")
        # I
        axi   = fig.add_axes([0.27,0.04,0.43,0.48], projection="hammer")
        caxi  = fig.add_axes([0.27,0.05,0.43,0.015])
        axi.tick_params(labelsize="x-small", labelcolor="0.3", )
        
        
        try: # python 3
            prop = {**dict(colorbar=colorbar, edgecolor="0.5", linewidth=0.5),**kwargs}
        except: #  python 2
            prop = dict(colorbar=colorbar, edgecolor="0.5", linewidth=0.5)
            for k,v in kwargs.items():
                prop[k] = v
                
        for i,ax_,cax_ in zip([1,2,3], [axg,axr,axi], [caxg,caxr,caxi]):
            if colored_by in ["visits", "density"]:
                field_val = {f:v for f,v in self.get_field_obsdensity(grid=grid, fid=[i]).items() if v>0}
            else:
                field_val = colored_by[i]
                
            self.show_fields(field_val, ax=ax_, cax=cax_, cmap=FIELDS_COLOR[i], **prop)
            
        return fig
    
    # =================== #
    #                     #
    # =================== #
    @property
    def _data(self):
        """ """
        return self.metatable if hasattr(self, "metatable") else self.data

    
class _ZTFDownloader_( object ):
    """ Virtual class that enable to download consistently ZTF data. 
    To use it, you need to inherite this and implement get_data_path()
    such that this method returns fullpath to the data given 
    `suffix` and `source` arguments.
    """
    def get_data_path(self, suffix=None, source=""):
        """ generic method to build the url/fullpath or the requested data.
        This method is based on the `builurl.py` module of ztfquery.

        **This is a virtual empty function ; inheriting class must implemented This**
        """
        raise NotImplementedError("the get_data_path() method must be implemented. ")

    
    # Generic that should automatically work as long as get_data_path is defined.
    def download_data(self, suffix=None, source="IRSA", download_dir=None,
                     show_progress = True, notebook=False, verbose=True,
                     nodl = False, overwrite=False, **kwargs):
        """ 
        Parameters
        ----------
        suffix: [string] -optional-
            What kind of data do you want?
            for science sources:
            - sciimg.fits (primary science image) *[used if suffix is None]*
            - mskimg.fits (bit-mask image)
            - psfcat.fits (PSF-fit photometry catalog)
            - sexcat.fits (nested-aperture photometry catalog)
            - sciimgdao.psf (spatially varying PSF estimate in DAOPhot's lookup table format)
            - sciimgdaopsfcent.fits (PSF estimate at science image center as a FITS image)
            - sciimlog.txt (log output from instrumental calibration pipeline)
            - scimrefdiffimg.fits.fz (difference image: science minus reference; fpack-compressed)
            - diffimgpsf.fits (PSF estimate for difference image as a FITS image)
            - diffimlog.txt (log output from image subtraction and extraction pipeline)
            - log.txt (overall system summary log from realtime pipeline)


        download_dir: [string] -optional-
            Directory where the file should be downloaded.
            If th
            
        overwrite: [bool] -optional-
            Check if the requested data already exist in the target download directory. 
            If so, this will skip the download except if overwrite is set to True.

            
        """
        from .io import download_single_url
        
        # Data Structure
        self._relative_data_path = self.get_data_path(suffix=suffix, source="None", **kwargs)
        
        # The IRSA location
        self.to_download_urls    = [buildurl._source_to_location_(source) + d_
                                     for d_ in self._relative_data_path]
        # Where do you want them?
        if download_dir is None: # Local IRSA structure
            self.download_location   = [buildurl._source_to_location_("local") + d_
                                        for d_ in self._relative_data_path]
            mkdir = True
        else:
            self.download_location   = [download_dir + "/%s%"%(d_.split("/")[-1])
                                        for d_ in self._relative_data_path]
            mkdir = False

        if nodl:
            return self.to_download_urls, self.download_location
            
        for url, fileout in zip(self.to_download_urls, self.download_location):
            if verbose: print(url)
            if not overwrite and os.path.isfile( fileout ):
                if verbose: print("%s already exists: skipped"%fileout)
                continue
            download_single_url(url,fileout=fileout, show_progress=show_progress,
                                    notebook=notebook, mkdir=mkdir)
        
    # --------- #
    #  GETTER   #
    # --------- #
    def get_local_data(self, suffix=None, exists=True):
        """ the lists of files stored in your local copy of the ztf database.
        [This methods uses the get_data_path() method assuming source='local']

        Parameters
        ----------
        suffix: [string] -optional-
            What kind of data do you want?
            for science sources:
            - sciimg.fits (primary science image) *[used if suffix is None]*
            - mskimg.fits (bit-mask image)
            - psfcat.fits (PSF-fit photometry catalog)
            - sexcat.fits (nested-aperture photometry catalog)
            - sciimgdao.psf (spatially varying PSF estimate in DAOPhot's lookup table format)
            - sciimgdaopsfcent.fits (PSF estimate at science image center as a FITS image)
            - sciimlog.txt (log output from instrumental calibration pipeline)
            - scimrefdiffimg.fits.fz (difference image: science minus reference; fpack-compressed)
            - diffimgpsf.fits (PSF estimate for difference image as a FITS image)
            - diffimlog.txt (log output from image subtraction and extraction pipeline)
            - log.txt (overall system summary log from realtime pipeline)

        exists: [bool] -optional-
            returns only the file that exists in your computer. 
            If false, this will return the expected path of the requested data, 
            even though they might not exist.

        Returns
        -------
        list
        """
        files = self.get_data_path(suffix=suffix, source="local")
        if not exists:
            return files
        return [f for f in files if os.path.isfile( f )]


    
class ZTFQuery( _ZTFTableHandler_, _ZTFDownloader_ ):
    """ """
    # ------------ #
    #  DOWNLOADER  #
    # ------------ #
    def load_metadata(self, kind="sci",
                        radec=None, size=None, mcen=None,
                        caltype=None,
                        sql_query=None, **kwargs):
        """ Querying for the metadata information that enables to reconstruct the URL to access the data.
        
        [This methods uses the .metasearch library, which is python wrapper of the the IRSA web API
        see https://irsa.ipac.caltech.edu/docs/program_interface/ztf_api.html]
        
        Parameters
        ----------
        kind: [str] -optional-
            What kind of data are you looking for:
            - sci : Science Exposures
            - raw : Raw Data
            - ref : Reference Images
            - cal : Bias or High Frequency Flat 
            any other entry will raise a ValueError

        // Generic Query

        sql_query: [None or string] -optional - 
            The where parameter can be set to a 'SQL WHERE' clause, with some restrictions. 
            [https://en.wikipedia.org/wiki/Where_(SQL)]
            Notably, function calls and sub-queries are not supported. You can use AND, OR, NOT, IN, BETWEEN, LIKE, IS, 
            the usual arithmetic and comparison operators, and literal values.
            Note that the where parameter is required in the absence of POS (a spatial constraint).
            WHERE clauses should be URL encoded [https://en.wikipedia.org/wiki/Query_string#URL_encoding].
            for instance  SPACE is encoded as '+' or "%20".
            If entry must be equal to a string, use `entry='str'` (with the quotes)
            Examples:
                get all the science field 600
                ```field=600```
                get all the science field 600 and having an airmass greater than 2
                ```field=600+AND+airmass>2```
                get all the science field 600 and having an airmass greater than 2 with a quadran ID been 1 or 3
                ```field=600+AND+airmass>2+AND+qid+IN+(1,3)```
                get observation taken since the 1st of Feb 2018 (julian date 2458150.5) with an airmass > 3
                ```airmass>3+AND+obsjd>2458150.5```
    
        // If not Calibration //

        ra,dec: [float/str]
            ICRS right ascension and declination in decimal degrees.
            It identifies the point which returned images must contain, or the center of the search region.

        size: [float/str/None] -optional-
            It consists of one or two (comma separated) values in decimal degrees. 
            (With POS=ra,dec)
            The first value is taken to be the full-width of the search region along the east axis at POS, 
            and the second is taken to be the full-height along the north axis. 
            Taken together, POS and SIZE define a convex spherical polygon on the sky with great circle edges - the search region. 
            During a query, this region is compared against the convex spherical polygons formed by connecting 
            the 4 corners of each image in a data-set to determine which images should be returned.
            
            If only one size value is specified, it is used as both the full-width and full-height.
            Negative sizes are illegal, and a width and height of zero indicate that the search region is a point.

        mcen: [bool] -optional-
            [If the size parameter is specified and non-zero, the mcen parameter is ignored] 

            The mcen parameter indicates that only the most centered image/image set 
            (with respect to POS) should be returned, rather than all images/image sets containing POS. 


        // If Calibration //
        caltype: [strin]
            which calibration type? 'bias' or 'hifreqflat'
            This classification will be added to the sql_query (caltype=`caltype`) 
            except if the sql_query already contains it. 
            If None, this will be ignored 
    
        """
        _test_kind_(kind)
        if kind not in ['cal']:
            # python3 -> self.metaquery = download_metadata(**{**locals(),**kwargs})
            self.metaquery = download_metadata(kind=kind, radec=radec, size=size, mcen=mcen, sql_query=sql_query, **kwargs)
        else:
            for k in ["radec", "size", "mcen"]:
                if locals()[k] is not None: warnings.warn("Calibration data requested, %s entry ignored"%k)
            if "caltype" not in sql_query and caltype is not None and caltype:
                sql_query = "caltype=%s"%caltype if sql_query is None else sql_query+"+AND+caltype='%s'"%caltype
                
            self.metaquery = download_metadata(kind=kind, sql_query=sql_query, **kwargs)
            
    
    def get_data_path(self, suffix=None, source=None):
        """ generic method to build the url/fullpath or the requested data.
        This method is based on the `builurl.py` module of ztfquery.
        
        Parameters
        ----------
        suffix: [string] -optional-
            What kind of data do you want?
            for science sources:
            - sciimg.fits (primary science image) *[used if suffix is None]*
            - mskimg.fits (bit-mask image)
            - psfcat.fits (PSF-fit photometry catalog)
            - sexcat.fits (nested-aperture photometry catalog)
            - sciimgdao.psf (spatially varying PSF estimate in DAOPhot's lookup table format)
            - sciimgdaopsfcent.fits (PSF estimate at science image center as a FITS image)
            - sciimlog.txt (log output from instrumental calibration pipeline)
            - scimrefdiffimg.fits.fz (difference image: science minus reference; fpack-compressed)
            - diffimgpsf.fits (PSF estimate for difference image as a FITS image)
            - diffimlog.txt (log output from image subtraction and extraction pipeline)
            - log.txt (overall system summary log from realtime pipeline)

        // if queried metadata is for kind calibration
            
        """
        # PIXELS
        if self.datakind in ['sci', "raw"]:
            
            filtercode,imgtypecode  = np.asarray(self.metatable[["filtercode","imgtypecode"]
                                                                    ].values.T, dtype="str")
            paddedfield      = np.asarray(["%06d"%f for f in self.metatable["field"].values],
                                              dtype="str")
            paddedccdid      = np.asarray(["%02d"%f for f in self.metatable["ccdid"].values],
                                              dtype="str")
            year, month, day, fracday = np.asarray([[l[:4],l[4:6],l[6:8],l[8:]]
                                for l in np.asarray(self.metatable["filefracday"].values,
                                              dtype="str") ]).T
  
            if self.datakind in ['sci']:
                qid  = np.asarray(self.metatable["qid"], dtype="str")
                # LIST of URL to download [SCIENCE]
                return  [buildurl.science_path(year_, month_, day_, fracday_, paddedfield_,
                                filtercode_, paddedccdid_, qid_,
                                imgtypecode=imgtypecode_, suffix=suffix, source=source)
                                
                            for year_, month_, day_, fracday_, paddedfield_, filtercode_,
                            paddedccdid_, qid_, imgtypecode_
                            in zip(year, month, day, fracday, paddedfield, filtercode,
                                       paddedccdid, qid, imgtypecode)]
            else:
                # LIST of URL to download [RAW]
                return  [buildurl.raw_path(year_, month_, day_, fracday_, paddedfield_,
                              filtercode_, paddedccdid_, 
                              imgtypecode=imgtypecode_, source=source)
                        for year_, month_, day_, fracday_, paddedfield_, filtercode_,
                        paddedccdid_,  imgtypecode_
                        in zip(year, month, day, fracday, paddedfield, filtercode,
                                   paddedccdid, imgtypecode)]
        # CALIBRATION
        elif self.datakind in ['cal']:
            year, month, day = np.asarray([[l[:4],l[4:6],l[6:]]
                                for l in np.asarray(self.metatable["nightdate"].values,
                                                        dtype="str") ]).T
            paddedccdid      = np.asarray(["%02d"%f for f in self.metatable["ccdid"].values],
                                              dtype="str")
            filtercode, qid,caltype  = np.asarray(self.metatable[["filtercode",
                                                                "qid","caltype"]].values.T,
                                                      dtype="str")
            # list of url to download [CAL]
            return  [buildurl.calibration_path(caltype_,
                                                                year_, month_, day_,
                                                                filtercode_, paddedccdid_, qid_,
                                                                suffix=suffix, source=source)
                                        for caltype_, year_, month_, day_, filtercode_, paddedccdid_, qid_
                                        in zip(caltype,year, month, day,filtercode, paddedccdid, qid) ]
        # PIXELS
        elif self.datakind in ['ref']:
            raise NotImplementedError("REFERENCE QUERYING NOT READY YET")

        
    # =============== #
    #  Properties     #
    # =============== #
    @property
    def datakind(self):
        """ """
        if not hasattr(self, "metaquery"):
            raise AttributeError("metaquery has not been loaded. Run load_metadata(). ")
        return self.metaquery.datakind
    
    @property
    def metatable(self):
        """ """
        if not hasattr(self, "metaquery"):
            raise AttributeError("metaquery has not been loaded. Run load_metadata(). ")
        return self.metaquery.metatable
        
#############################
#                           #
#  Addition Queries         #
#                           #
#############################
_NIGHT_SUMMARY_URL = "http://www.astro.caltech.edu/~tb/ztfops/sky/"
def download_night_summary(night):
    """ 
    Parameters
    ----------
    night: [string]
        Format: YYYYMMDD like for instance 20180429
    """
    import requests
    from pandas import DataFrame
    from .io import _load_id_
    summary = requests.get(_NIGHT_SUMMARY_URL+"%s/exp.%s.tbl"%(night,night),
                               auth=_load_id_("ztfops", askit=True)
                          ).content.decode('utf-8').splitlines()
    columns = [l.replace(" ","") for l in summary[0].split('|') if len(l.replace(" ",""))>0]
    data    = [l.split() for l in summary[1:] if not l.startswith('|') and len(l)>1]
    dataf   = DataFrame(data=data, columns=[l if l!= "fil" else "fid" for l in columns])
    dataf["fid"][dataf["fid"]=="4"] = "3"
    return dataf


class NightSummary( _ZTFTableHandler_, _ZTFDownloader_ ):
    def __init__(self, night):
        """ """
        self.night = night
        
        self.data_all  = download_night_summary(night)
        
        self.data  = self.data_all[self.data_all["type"]=="targ"]
        
    # ================ #
    #  Methods         #
    # ================ #
    # --------- #
    #  GETTER   #
    # --------- #
    def get_observed_information(self, obstype="targ", columns=["field","ra","dec"]):
        """ get a DataFrame (pandas) of the requested columns for the given obstype. 

        Parameters
        ----------
        obstype: [string]
            Type of observation. 
            Could be: 'bias', 'dark', 'flat', or 'targ'
            
        columns: [string or list of]
            Any field available in data (check the list by doing THIS.data.columns)

        Returns
        -------
        DataFrame
        """
        return self.data[self.data['type']==obstype][columns]

    # Download Data
    def set_metadata(self, kind, **kwargs):
        """ Set the mate information get_data_path need
        
        Important: Some kwargs are mandatory dependending of you given kind:

        - for kind "sci":  ["paddedccdid", "qid"]
        - for kind "raw":  ["paddedccdid"]
        (other kind not implemented yet)
        """
        self._metadata = {}

        MANDATORY = {"sci":["paddedccdid", "qid"],
                     "raw":["paddedccdid"],
                     "ref":{},
                     "cal":{}
                     }
            
        DEFAULT = {"sci":{"imgtypecode":"o"},
                   "raw":{"imgtypecode":"o"},
                    "ref":{},
                   "cal":{}}
            
        if kind not in ["raw","sci"]:
            raise NotImplementedError("Only Science ('sci') or Raw ('raw') kinds ready (%s given)."%kind)
        # Requested input
        for k in MANDATORY[kind]:
            if k not in kwargs.keys():
                raise ValueError("%s should be provided for kind: %s"%(k,kind))
            
        # -> python3    self._metadata = **{DEFAULT[kind], **kwargs}
        
        self._metadata = DEFAULT[kind]
        self._metadata["kind"] = kind
        for k,v in kwargs.items():
            self._metadata[k] = v
        # -> python3    self._metadata = **{DEFAULT[kind], **kwargs}; self._metadata["kind"] = kind
            
    # WRONG SO FAR
    def get_data_path(self, mask=None, suffix=None, source=None, verbose=False, ):
        """ generic method to build the url/fullpath or the requested data.
        This method is based on the `builurl.py` module of ztfquery.
        
        Parameters
        ----------
        mask: [None / list of int / boolean array] -optional-
           only use the data entry for the given mask:
           ```
           fileroots = np.asarray(self.data["fileroot"])
           if mask is not None:
               fileroots = fileroots[mask]
            ```

        suffix: [string] -optional-
            What kind of data do you want?
            for science sources:
            - sciimg.fits (primary science image) *[used if suffix is None]*
            - mskimg.fits (bit-mask image)
            - psfcat.fits (PSF-fit photometry catalog)
            - sexcat.fits (nested-aperture photometry catalog)
            - sciimgdao.psf (spatially varying PSF estimate in DAOPhot's lookup table format)
            - sciimgdaopsfcent.fits (PSF estimate at science image center as a FITS image)
            - sciimlog.txt (log output from instrumental calibration pipeline)
            - scimrefdiffimg.fits.fz (difference image: science minus reference; fpack-compressed)
            - diffimgpsf.fits (PSF estimate for difference image as a FITS image)
            - diffimlog.txt (log output from image subtraction and extraction pipeline)
            - log.txt (overall system summary log from realtime pipeline)
            
            [ignored for raw data, i.e. no suffixes as there is only 1 data available]

        // if queried metadata is for kind calibration
            
        """
        if not hasattr(self,"_metadata"):
            raise AttributeError("you did not set the metadata, you must (see self.set_metadata()) ")

        fileroots = np.asarray(self.data["fileroot"])
        if mask is not None:
            fileroots = fileroots[mask]
        
        
        # Science Products
        if self._metadata["kind"] in ['sci']:
            from .buildurl import fileroot_to_science_url
            if suffix is None:
                suffix = "sciimg.fits"

            return [fileroot_to_science_url(fileroot, self._metadata["paddedccdid"], self._metadata["qid"],
                            imgtypecode=self._metadata["imgtypecode"],
                            suffix=suffix, source=source,
                            verbose=verbose)
                        for fileroot in fileroots]
        # Raw Data
        elif self._metadata["kind"] in ["raw"]:
            from .buildurl import fileroot_to_raw_url
            return [fileroot_to_raw_url(fileroot, self._metadata["paddedccdid"],
                        imgtypecode=self._metadata["imgtypecode"],
                                            source=source, verbose=verbose)
                        for fileroot in fileroots]
        else:
            raise NotImplementedError("Only Science ('sci') or Raw ('raw') kinds ready (%s given)."%kind)

    



#!/usr/bin/env python
#

""" Alert Reader and basic tools """


import numpy as np
from .lightcurve import FILTER_COLORS, FILTER_CODE


def display_alert(alert, savefile=None, show_ps_stamp=False):
    """ """
    if type(alert) == str:
        # This is a file:
        al = AlertReader.load(alert)
    else:
        al = AlertReader(alert)

    return al.show(savefile, show_ps_stamp=show_ps_stamp)



def query_alert(alertid):
    """ """
    print("This service will soon be added")
    pass
####################
#                  #
#  Basic Class     #
#                  #
####################
class AlertReader():
    """ """    
    def __init__(self, alert):
        """ Input: Avro alert loaded. If you have a avrofile, use .load()"""
        self.alert = alert

        
    @classmethod
    def load(cls, avrofile):
        """ """
        import fastavro
        with open(avrofile, "rb") as fo:
            av = next(fastavro.reader(fo), None)
        return cls(av)
    

    # ================ #
    #   Methods        #
    # ================ #
    def get_stamp(self, which):
        """ return the HDUImage stamps for `which`.
        where `which` is one of the following ['Science','Template','Difference']
        """
        import gzip, io
        from astropy.io import fits
        with gzip.open(io.BytesIO(self.alert.get("cutout%s"%which)["stampData"]), 'rb') as f:
            return fits.open(io.BytesIO(f.read()))[0]

    # --------- #
    #  GETTER   #
    # --------- #
    def get_history_photopoints(self):
        """ list of alert history having been detected (not upper limit, real photometric points) """
        return [d for d in self.alert['prv_candidates'] if d.get('candid') is not None]

    def get_history_upperlimits(self):
        """ list of alert history not  detected (upper limit, not real photometric points) """
        return [d for d in self.alert['prv_candidates'] if d.get('candid') is None]

    def download_ps_stamp(self, size=240, color=["y","g","i"]):
        """ Downloading the PS colored cutout stamp centered at the alert candidate location
        Returns the PIL Image
         = uses `get_ps_stamp` from ztfquery.utils.stamps.
        """
        from .utils.stamps import get_ps_stamp
        return get_ps_stamp(self.alert["candidate"]["ra"], self.alert["candidate"]["dec"],  size=size, color=color)
    
    # --------- #
    # PLOTTER   #
    # --------- #
    def show(self, savefile=None, show_ps_stamp=False):
        """ """
        from astropy import visualization
        from astropy.time import Time
        import matplotlib.dates as mdates
        from matplotlib.colors import Normalize
        import matplotlib.pyplot as mpl
    
        # ----------- #
        # Global      #
        # ----------- #
        prop = dict(marker="o", mec="0.7", ms=8, ecolor="0.7", ls="None")
    
        # ----------- #
        #   Methods   #
        # ----------- #
        def show_fid_lc(ax):
            """ """
            if len(self.get_history_photopoints()) == 0:
                return
            mag,magerr, jd, fid = np.asarray([ [d[k] for k in ["magpsf","sigmapsf","jd","fid"]]
                                                   for d in self.get_history_photopoints()]).T
            for j,i in enumerate([1,2,3]):
                if i in fid:
                    flag_fid = fid==i
                    ax.errorbar([Time(jd_, format="jd").datetime for jd_ in jd[flag_fid]], 
                                    mag[flag_fid], yerr= magerr[flag_fid], 
                                    label="magpsf %s"%FILTER_CODE[j], mfc=FILTER_COLORS[j], **prop)
            
        def show_fid_uplim(ax):
            """ """
            if len(self.get_history_upperlimits()) == 0:
                return
            upmag, jdup, fidup    = np.asarray([ [d[k] for k in ["diffmaglim","jd","fid"]]
                                                     for d in self.get_history_upperlimits()]).T
            for j,i in enumerate([1,2,3]):
                if i in fidup:
                    flag_fid = fidup==i
                    ax.errorbar([Time(jd_, format="jd").datetime for jd_ in jdup[flag_fid]], 
                                upmag[flag_fid], yerr=0.2, lolims=True,
                                    color=FILTER_COLORS[j], ls="None", 
                                    label="_no_legend_")
                
        # ----------- #
        #   Axes      #
        # ----------- #
        fig = mpl.figure(figsize=[9,5])

        ref, width, heigh = 0.1,0.15, 0.25
        ypos, span  = 0.65, 0.05
        # Stamps
        aximg = fig.add_axes([ref,ypos,width,heigh])
        axref = fig.add_axes([ref+(width+span),ypos,width,heigh])
        axdif = fig.add_axes([ref+(width+span)*2,ypos,width,heigh])
        if show_ps_stamp:
            axps   = fig.add_axes([ref+(width+span*1.5)*3,ypos,width,heigh])
            axps.set_yticks([])
            axps.set_xticks([])
        # - LC
        axlc   = fig.add_axes([ref, 0.1,(width+span)*2+width,0.5])
        axlc.set_xlabel("Date", fontsize="large")
        axlc.set_ylabel("mag (magpsf)", fontsize="large")
    
        # ----------------- #
        #  Plotting Stamps  #
        # ----------------- # 
        # Loop Over the stamps
        for ax_, source in zip([aximg, axref, axdif], ['Science','Template','Difference']):
            data_ = visualization.AsinhStretch(self.get_stamp(source).data).a
            ax_.imshow(data_, norm=Normalize(*np.percentile(data_[data_==data_], [0.5,99.5])), aspect="auto")
            ax_.set_yticks([])
            ax_.set_xticks([])
            ax_.set_title(source)

        if show_ps_stamp:
            try:
                img = self.download_ps_stamp(color=["y","g","i"])
                axps.imshow(np.asarray(img))
            except:
                print("Pan-STARRS stamp failed.")
            axps.set_title("Pan-STARRS (y/g/i)")
        # ----------- #
        #  History    #
        # ----------- #
        show_fid_lc(axlc)
        axlc.invert_yaxis()
        show_fid_uplim(axlc)
        axlc.legend(loc="best")
    
        # ----------- #
        #  Alert      #
        # ----------- #
        prop['marker'] = "D"
        prop['mec'] = FILTER_COLORS[self.alert["candidate"]["fid"]-1]
        axlc.errorbar(Time(self.alert["candidate"]["jd"], format="jd").datetime, self.alert["candidate"]["magpsf"], 
                    yerr= self.alert["candidate"]["sigmapsf"],
                    mfc=FILTER_COLORS[self.alert["candidate"]["fid"]-1],label="_no_legend_", **prop)
    
        # add text
        info = []
        for k in ["rb","fwhm","nbad", "elong", "isdiffpos"]:
            try:
                info.append("%s : %.3f"%(k,self.alert["candidate"].get(k)) )
            except:
                info.append("%s : %s"%(k,self.alert["candidate"].get(k)) )
            
        for kk in ["objectidps", "sgscore", "distpsnr","srmag"]:
            for k in [k for k in self.alert["candidate"].keys() if kk in k]:
                info.append("%s : %s"%(k,self.alert["candidate"].get(k)) )

        fig.text(0.68,0.6, " \n".join(info), va="top", fontsize="medium", color="0.4")
        fig.text(0.005,0.995, "alert: ID: %s (RA: %.5f | Dec: %.5f | Filter: %s)"%(self.alert.get("candid"),
                                                                                   self.alert['candidate']['ra'],self.alert['candidate']['dec'],
                                                                                   FILTER_CODE[self.alert['candidate']['fid']-1]),
                     fontsize="medium", color="k", va="top", ha="left")
        axlc.xaxis.set_major_formatter(mdates.DateFormatter('%m/%d/%y'))

        if savefile is not None:
            fig.savefig(savefile, dpi=250)
            
        return fig



#! /usr/bin/env python
#

"""  Access SEDM data from pharos """


PHAROS_BASEURL = "http://pharos.caltech.edu"
import os
import requests
import json
import numpy as np
import pandas
import warnings
from . import io

SEDMLOCAL_BASESOURCE = io.LOCALSOURCE+"SEDM"
SEDMLOCALSOURCE = SEDMLOCAL_BASESOURCE+"/redux"
if not os.path.exists(SEDMLOCAL_BASESOURCE):
    os.makedirs(SEDMLOCAL_BASESOURCE)
if not os.path.exists(SEDMLOCALSOURCE):
    os.makedirs(SEDMLOCALSOURCE)
    

#######################
#                     #
#  High level method  #
#                     #
#######################
def _download_sedm_data_(night, pharosfile, fileout=None, verbose=False):
    """ """
    url = PHAROS_BASEURL+"/data/%s/"%night+pharosfile
    if verbose:
        print(url)
    return io.download_single_url(url,fileout=fileout,
                                  auth=io._load_id_("pharos"),
                                  cookies="no_cookies")

def _relative_to_source_(relative_datapath, source=None):
    """ """
    if source is None:
        return relative_datapath
    if source in ["pharos"]:
        return [PHAROS_BASEURL+"/data/"+l for l in relative_datapath]
    if source in ["local"]:
        return [SEDMLOCALSOURCE+"/"+l for l in relative_datapath]
    
def get_night_file(night):
    """ get the what.list for a given night 
    night format: YYYYMMDD 
    """
    response = _download_sedm_data_(night, "what.list")
    return response.text.splitlines()


def get_pharos_night_data(date, auth=None):
    """ """
    username,password = io._load_id_("pharos") if auth is None else auth
    requests_prop = {"data":json.dumps({"obsdate":date,
                                            "username":username,
                                            "password":password,
                                            }),
                         "headers":{'content-type': 'application/json'}}
            
    t = requests.post(PHAROS_BASEURL+"/get_user_observations", **requests_prop).text
    if "data" not in t:
        raise IOError("night file download fails. Check you authentification maybe?")
    return np.sort(json.loads(t)["data"])
#######################
#                     #
#  INTERNAL JSON DB   #
#                     #
#######################
# 20181012 20181105
EMPTY_WHAT_DF = pandas.DataFrame(columns=["filename","airmass", "shutter", "exptime", "target", "night"])

def _parse_line_(line):
    """ """
    try:
        filename, rest = line.split('(')
        info, what = rest.split(")")
        what = what.replace(":", "")
        return [filename.replace(" ","")]+info.split("/")+[what.replace(" [A]","").strip()]
    except:
        return None

    
def whatfiles_to_dataframe(whatfile):
    """ """
    parsed_lines = [_parse_line_(l_) for l_ in whatfile]
    
    return pandas.DataFrame([l for l in parsed_lines if l is not None],
                                columns=["filename","airmass", "shutter", "exptime", "target"])

class _SEDMFiles_():
    """ """
    SOURCEFILE  = SEDMLOCAL_BASESOURCE+"/whatfiles.json"
    PHAROSFILES = SEDMLOCAL_BASESOURCE+"/pharosfiles.json"
    def __init__(self):
        """ """
        self.load()

    def get_night_data(self, night, from_dict=False):
        """ """
        if from_dict:
            df_night = pandas.DataFrame(whatfiles_to_dataframe(self._data[night]))
            df_night["night"] = night
            return df_night
        return self.data[self.data["night"].isin(np.atleast_1d(night))]

    def get_pharos_night_data(self, night):
        """ """
        return self._pharoslist[night]
        
    def get_data_betweenrange(self, start="2018-08-01", end=None):
        """ """
        lower_bound = True if start is None else (self.datetime>start)
        upper_bound = True if end is None and end not in ["now","today"] else (self.datetime<end)
        return self.data[lower_bound & upper_bound]

    def get_target_data(self, target, timerange=None):
        """ """
        data_ = self.data if timerange is None else self.get_data_betweenrange(*timerange)
        return data_[data_["target"].isin(np.atleast_1d(target))]

    def get_observed_targets(self, timerange=None):
        """ """
        data_ = self.data if timerange is None else self.get_data_betweenrange(*timerange)
        return np.unique(data_["target"])
    
    def get_nights_with_target(self, target, timerange=None):
        """ """
        return np.unique( self.get_target_data(target, timerange=timerange)["night"] )
    
    # -------- #
    #    I/O   #
    # -------- # 
    def download_nightrange(self, start="2018-08-01", end="now", update=False, pharosfiles=False, dump=True):
        """ """
        if end is None or end in ["today", "now"]:
            from datetime import datetime 
            today = datetime.today()
            end   = today.isoformat().split("T")[0]
            
        self.add_night(["%4d%02d%02d"%(tt.year,tt.month, tt.day) for tt in pandas.date_range(start=start, end=end) ], update=update)
        if pharosfiles:
            self.add_pharoslist(["%4d%02d%02d"%(tt.year,tt.month, tt.day) for tt in pandas.date_range(start=start, end=end) ], update=update)

        if dump:
            self.dump("whatfile" if not pharosfiles else "both")
            
    def add_night(self, night, update=False):
        """ night (or list of) with the given format YYYYMMDD 
        if the given night is already known, this will the download except if update is True 
        """
        for night_ in np.atleast_1d(night):
            if night_ in self._data and not update:
                continue
            self._data[night_] = get_night_file(night_)
            
        self.dump("whatfile")
        self._build_dataframe_()
        
    def load(self):
        """ """
        # What Files
        if os.path.isfile( self.SOURCEFILE ):
            self._data = json.load( open(self.SOURCEFILE, 'r') )
        else:
            self._data = {}
        # What Pharos Data
        if os.path.isfile( self.PHAROSFILES ):
            self._pharoslist = json.load( open(self.PHAROSFILES, 'r') )
        else:
            self._pharoslist = {}
            
        self._build_dataframe_()
        
    def dump(self, which="both"):
        """ Save the current version of whatfiles and or pharos files on your computer.

        Parameters
        ----------
        which: [str] -optional-
            what kind of data do you want to dump ?
            - whatfile
            - pharosfile
            - both
        """
        if not which in ["whatfile","pharosfile","both","*", "all"]:
            raise ValueError("which can only be whatfile or pharosfile or both")
        
        if which in ["whatfile", "both","*", "all"]:
            with open(self.SOURCEFILE, 'w') as outfile:
                json.dump(self._data, outfile)
                
        if which in ["pharosfile","both","*", "all"]:
            with open(self.PHAROSFILES, 'w') as outfile:
                json.dump(self._pharoslist, outfile)
        
    
    def _build_dataframe_(self):
        """ """
        if len(self._data.keys())>0:
            self.data = pandas.concat(self.get_night_data(night, from_dict=True) for night in self._data.keys())
        else:
            self.data = EMPTY_WHAT_DF

    # ---------------- #
    #  Pharos Data     #
    # ---------------- #
    def add_pharoslist(self, night, update=False):
        """ """
        for night_ in np.atleast_1d(night):
            if night_ in self._pharoslist and not update:
                continue
            try:
                self._pharoslist[night_] = [l.replace("/data/","") for l in get_pharos_night_data(night_)]
            except:
                warnings.warn("Pharos List download: Failed for %s"%night_)
            
        self.dump("pharosfile")

    # ================ #
    #   Properties     #
    # ================ #
    @property
    def datetime(self):
        """ pandas.to_datetime(p.sedmwhatfiles.data["night"]) """
        return pandas.to_datetime(self.data["night"])
    
##################
#                #
#  PHAROS        #
#                #
##################
class SEDMQuery( object ):
    """ """
    PROPERTIES = ["auth", "date"]
    def __init__(self, auth=None, date=None):
        """ """
        self.sedmwhatfiles = _SEDMFiles_()
        self.reset()
        self.set_date(date)
        self.set_auth(io._load_id_("pharos") if auth is None else auth)
        
    def reset(self):
        """ set the authentification, date and any other properties to default """
        self._properties = {k:None for k in self.PROPERTIES}
        
    # -------- #
    #  SETTER  #
    # -------- #
    def set_date(self, date):
        """ attach a date for faster night access interation """
        self._properties["date"] = date
    
    def set_auth(self, auth):
        """ provide your authentification. """
        self._properties["auth"] = auth

    # ----------- #
    # Downloader  #
    # ----------- #
    def download_night_fluxcal(self, night, nodl=False, auth=None, download_dir="default",
                                 show_progress=False, notebook=False, verbose=True,
                                 overwrite=False, nprocess=None):
        """ download SEDM fluxcalibration file for the given night
        
        Parameters
        ----------
        nodl: [bool] -optional-
            do not launch the download, instead, returns 
            list of queried url and where they are going to be stored.
            
        download_dir: [string] -optional-
            Directory where the file should be downloaded.
            If th
            
        overwrite: [bool] -optional-
            Check if the requested data already exist in the target download directory. 
            If so, this will skip the download except if overwrite is set to True.

        nprocess: [None/int] -optional-
            Number of parallel downloading you want to do. 
            If None, it will be set to 1 and will not use multiprocess

        auth: [str, str] -optional-
            [username, password] of you pharos account.
            If used, information stored in ~/.ztfquery will be ignored.
            
        Returns
        -------
        Void or list (see nodl)
        """
        relative_path = [l for l in self.get_night_data(night, source='pharos') if l.split("/")[-1].startswith("fluxcal")]
        return self._download_from_relative_path_(relative_path, nodl=nodl, auth=auth, download_dir=download_dir,
                                          show_progress=show_progress, notebook=notebook, verbose=verbose,
                                          overwrite=overwrite, nprocess=nprocess)

    def download_target_data(self, target, which="cube", extension="fits",
                                 timerange=["2018-08-01", None],
                                 nodl=False, auth=None, download_dir="default",
                                 show_progress=False, notebook=False, verbose=True,
                                 overwrite=False, nprocess=None ):
        """ 
        download SEDM data associated to the given target. 
        
        Parameters
        ----------
        target: [string] 
            Name of a source (e.g. ZTF18abuhzfc) of any part of a filename (i.e. 20180913_06_28_51)
            
        which: [string] -optional-
            kind oif data you want. 
            - cube / spec / ccd / all

        extension: [string] -optional-
            Extension of the file 
            - these exist depending on the file you want: fits / png / pdf / pkl / all

        timerange: [iso format dates] -optional-
            time range between which you are looking for file.
            If the dates are not yet stored in you whatfiles.json, this will first download it.
            if the second data is None, it means 'today'

        nodl: [bool] -optional-
            do not launch the download, instead, returns 
            list of queried url and where they are going to be stored.
            
        download_dir: [string] -optional-
            Directory where the file should be downloaded.
            If th
            
        overwrite: [bool] -optional-
            Check if the requested data already exist in the target download directory. 
            If so, this will skip the download except if overwrite is set to True.

        nprocess: [None/int] -optional-
            Number of parallel downloading you want to do. 
            If None, it will be set to 1 and will not use multiprocess

        auth: [str, str] -optional-
            [username, password] of you pharos account.
            If used, information stored in ~/.ztfquery will be ignored.
            
        Returns
        -------
        Void or list (see nodl)
        """
        # Build the path (local and url)
        if "astrom" in which or "guider" in which:
            print("TMP which=astrom fixe")
            relative_path = [l.replace("e3d","guider").replace(target,"astrom") for l in self.get_data_path(target, which="cube",extension="fits", timerange=timerange, source="pharos")]
        else:
            relative_path = self.get_data_path(target, which=which,extension=extension, timerange=timerange, source="pharos")
            
        return self._download_from_relative_path_(relative_path, nodl=nodl, auth=auth, download_dir=download_dir,
                                          show_progress=show_progress, notebook=notebook, verbose=verbose,
                                          overwrite=overwrite, nprocess=nprocess)
                                               
    # - Internal method
    def _download_from_relative_path_(self, relative_path,
                                          nodl=False, auth=None, download_dir="default",
                                          show_progress=False, notebook=False, verbose=True,
                                          overwrite=False, nprocess=None):
        """ Given a relative path, this builds the data to download and where to.

        Parameters
        ----------
        nodl: [bool] -optional-
            do not launch the download, instead, returns 
            list of queried url and where they are going to be stored.
            
        download_dir: [string] -optional-
            Directory where the file should be downloaded.
            If th
            
        overwrite: [bool] -optional-
            Check if the requested data already exist in the target download directory. 
            If so, this will skip the download except if overwrite is set to True.

        nprocess: [None/int] -optional-
            Number of parallel downloading you want to do. 
            If None, it will be set to 1 and will not use multiprocess

        auth: [str, str] -optional-
            [username, password] of you pharos account.
            If used, information stored in ~/.ztfquery will be ignored.
            
        Returns
        -------
        Void or list (see nodl)
        """
        self.to_download_urls  = _relative_to_source_(relative_path, "pharos")
        if download_dir is None or download_dir in ["default"]:
            self.download_location = _relative_to_source_(relative_path, "local")
        else:
            self.download_location = [download_dir+f.split("/")[-1] for f in self.to_download_urls]
            
        if not nodl:
            io.download_url(self.to_download_urls, self.download_location,
                        show_progress = show_progress, notebook=notebook, verbose=verbose,
                        overwrite=overwrite, nprocess=nprocess,
                        auth=self._properties["auth"] if auth is None else auth)
        
        return self.to_download_urls, self.download_location
    # -------- #
    #  GETTER  # 
    # -------- #
    def get_data_path(self, target, which="cube", extension="fits", source="pharos", timerange=["2018-08-01", None]):
        """ get the datapath for the given target. 
        this is used to build the url that will be queried (see, download_target_data) and the look 
        for file in your computer (see, get_local_data)
        
        Parameters
        ----------
        target: [string] 
            Name of a source (e.g. ZTF18abuhzfc) of any part of a filename (i.e. 20180913_06_28_51)
            
        which: [string] -optional-
            kind oif data you want. 
            - cube / spec / ccd / all

        extension: [string] -optional-
            Extension of the file 
            - these exist depending on the file you want: fits / png / pdf / pkl / all

        timerange: [iso format dates] -optional-
            time range between which you are looking for file.
            If the dates are not yet stored in you whatfiles.json, this will first download it.
            if the second data is None, it means 'today'
            
        source: [string] -optional-
            Where are you looking for data.
            - pharos (online)
            - local (your computer)
            
        Returns
        -------
        list of Path
        """
        all_data = []
        for night, fileid in self.get_target_data(target, timerange=timerange)[["night","filename"]].values:
            all_data+=[l for l in self.get_night_data(night, source=source) if fileid.split(".")[0] in l
                               and (which in ["*", "all"] or 
                                   ((which in ["cube"] or "cube" in which) and "/e3d" in l) or
                                   ((which in ["spec"] or "spec" in which) and "/spec_" in l) or
                                   ((which in ["ccd"] or "ccd" in which) and "/crr_" in l) or
                                   ((which in ["wcs","guider","astrom"] or "guider" in which) and "/guider_" in l and "astrom" in which)
                                    )
                                and (extension in ["*", "all"] or extension in l or l.split(".")[-1] in extension)
                               ]
            
        return all_data
    
    def get_local_data(self, target, which="cube", extension="fits", **kwargs):
        """ get existing for in you computer associated to the given target

        Parameters
        ----------
        target: [string] 
            Name of a source (e.g. ZTF18abuhzfc) of any part of a filename (i.e. 20180913_06_28_51)
            
        which: [string] -optional-
            kind oif data you want. 
            - cube / spec / ccd / all

        extension: [string] -optional-
            Extension of the file 
            - these exist depending on the file you want: fits / png / pdf / pkl / all

        kwargs goes to get_data_path()

        Returns
        -------
        full path
        """
        return self.get_data_path(target, which=which, extension=extension, source="local", **kwargs)

    def get_target_data(self, target, timerange=None):
        """ dictionary containing the dates and file id corresponding to the given target.
        this is based on the whatfiles.json stored in your computer under the SEDM directory
        
        Parameters
        ----------
        target: [string] 
            Name of a source (e.g. ZTF18abuhzfc) of any part of a filename (i.e. 20180913_06_28_51)

        timerange: [iso format dates / None] -optional-
            time range between which you are looking for file.
            If the dates are not yet stored in you whatfiles.json, this will first download it.
            if the second data is None, it means 'today'
            - 
            if None the instance timerange is not updated.

        Returns
        -------
        dict {date:[list of fileid ],...}
        """
        if timerange is not None:
            self.update_sedmdata(timerange)

        return self.sedmwhatfiles.get_target_data(target, timerange=timerange)


    def get_standards(self, timerange=None, use="*"):
        """ """
        std_names = np.unique([k for k in self.sedmwhatfiles.get_observed_targets(timerange) if 'STD' in k and (use in ["*","all"] or k in use)])
        return self.get_target_data(std_names, timerange)
        
        
    def update_sedmdata(self, timerange=["2018-08-01", None], pharosfiles=False, dump=True, **kwargs):
        """ update the local SEDm whatfiles 
        
        Parameters
        ----------
        timerange: [iso format dates] -optional-
            time range between which you are looking for file.
            If the dates are not yet stored in you whatfiles.json, this will first download it.
            if the second data is None, it means 'today'

        pharosfiles: [bool] -optional-
            Do you also what to get the list of files accessible from pharos for the given timerange ?

        dump: [bool] -optional-
            Once all what is requested is downloaded, shall the local file be updated too ?

        **kwargs goes to download_nightrange()

        """
        self.sedmwhatfiles.download_nightrange(*timerange, pharosfiles=pharosfiles, dump=dump, **kwargs)
        
    # = Get Night Data = #
    def get_night_data(self, date, source="pharos"):
        """  get all the data of the given date you have access to:
        
        Parameters
        ----------
        date: [string]
           format YYYYMMDD

        source: [string] -optional-
            Where are you looking for data.
            - pharos (online, only the one you have access to)
            - local (your computer,  only the one you have already downloaded)
        
        Returns
        -------
        list of file
        """
        if source in ["pharos", "sedm"]:
            if date not in self.sedmwhatfiles._pharoslist.keys():
                self.sedmwhatfiles.add_pharoslist(date, update=True)
            return self.sedmwhatfiles.get_pharos_night_data(date)
        
        if source in ["what"]:
            if date not in self.sedmwhatfiles.data["night"]:
                self.sedmwhatfiles.add_night(date)
            return self.sedmwhatfiles.data[self.sedmwhatfiles.data["night"]==date]
            
        elif source in ["local"]:
            return self._get_local_night_data_(date)
        raise ValueError("unknown source: %s, use 'local' or 'pharos'"%source)

    def _get_local_night_data_(self, date):
        """ """
        date_dir = SEDMLOCALSOURCE+"/%s/"%date
        if not os.path.exists( date_dir ):
            return []
        return [date_dir+l for l in os.listdir( date_dir )]
    # ============== #
    #  Properties    #
    # ============== #
    @property
    def date(self):
        """ """
        return self._properties["date"]

    @property
    def night_files(self):
        """ """
        if not hasattr(self, "_night_files") or self._night_files is None:
            self._night_files = self.get_night_data(self.date, source="pharos")
            
        return self._night_files
    
